// Written by Epsi Sayidina
// 24 June 1999
// 0493020349 - Mesin UI
// West Java, Indonesia

unit Matrix;

interface

Uses SysUtils, Classes;

Type
  Float=Real;

  TMatrix = class
  Private
    FArr: array of array of Float;
    Fm, Fn: byte;
    Function GetValue(i, j: byte): Float;
    Procedure SetValue(i, j: byte; const NewValue: Float);
  Public
    Constructor Create(m, n: byte);
    Procedure AssignFloat(const AMatrix: Array of Float);
    Property Value[i, j: byte]: Float read GetValue write SetValue; default;
    {not a zero based index}
  end;

  TSquareMatrix = class(TMatrix)
    Constructor Create(size: byte);
    Function Minor(i, j: byte): TSquareMatrix;
    Function CoFactor(i, j: byte): Float;
    Function Determinant: Float;     {Using cramer method}
  end;

  TCustomTrend = class
  // Referensi: [19]
  // John E. Biegel, "Pengendalian Produksi"
  // Akademika Pressindo, Jakarta 1992
  // halaman 28-36
  Private
    Function Sum_t: integer;
    Function Sum_t2: integer;
    Function GetValue(t: integer): Float; virtual; abstract;
  Protected
    a: Float;
    FArr: Array of Float;
    f: byte; {lost degrees of freedom}
  Public
    Function Equation: String; virtual; abstract;
    Function Total: Float;
    Function Count: Integer;
    Function StandardDeviation: Float;
    Property Value[t: integer]: Float read GetValue; default;
    Property n: Integer Read Count;
  end;

  TConstantTrend = class(TCustomTrend)
  Private
    Function GetValue(t: integer): Float; override;
  Public
    Function Equation: String; override;
    Constructor Create(Var AnArr: Array of Float); virtual;
  end;

  TLinearTrend = class(TCustomTrend) {Least Square Method}
  Protected
    b: Float;
  Private
    Function GetValue(t: integer): Float; override;
  Public
    Constructor Create(Var AnArr: Array of Float); virtual;
    Function Equation: String; override;
  end;

  TCustomCyclicTrend = class(TCustomTrend)
  Protected
    u, v: Float;
    nc : integer;  // Period per cyclic
  Public
    Procedure SetNumberOfCyclic(NewNumberOfCyclic:integer);
  end;

  TCyclicTrend = class(TCustomCyclicTrend)
  Private
    Function GetValue(t: integer): Float; override;
  Public
    Constructor Create
    (Var AnArr: Array of Float; Const NumberOfCyclic : integer); virtual;
    Function Equation: String; override;
  end;

  TLinearCyclicTrend = class(TCustomCyclicTrend)
  Private
    b : Float;
    Function GetValue(t: integer): Float; override;
  Public
    Constructor Create
    (Var AnArr: Array of Float; Const NumberOfCyclic : integer); virtual;
    Function Equation: String; override;
  end;

implementation

Function FToStr(F: Float): String;
begin
  if F>0 then Result:='+ '+FloatToStrF(F, ffFixed, 7,3)
  else Result:='- '+FloatToStrF(Abs(F), ffFixed, 7,3);
end;

Constructor TMatrix.Create(m, n: byte);
begin
  inherited create;
  Fm:=m;
  Fn:=n;
  SetLength(FArr, m, n);
end;

Function TMatrix.GetValue(i, j: byte): Float;
begin
  Result:=FArr[i-1, j-1];
end;

Procedure TMatrix.SetValue(i, j: byte; const NewValue: Float);
begin
  FArr[i-1, j-1]:= NewValue;
end;

Procedure TMatrix.AssignFloat(const AMatrix: Array of Float);
var i, j: byte;
begin
  if High(AMatrix)=Fm*Fn-1 then For i:=1 to Fm do For j:=1 to Fn do
     FArr [i-1, j-1]:= AMatrix[(i-1)*Fn+(j-1)];
end;

Constructor TSquareMatrix.Create(Size: byte);
begin
  inherited create(Size, Size);
end;

Function TSquareMatrix.Minor(i, j: byte): TSquareMatrix;
var Temp: TSquareMatrix;
    ei, ej, fi, fj: byte;
begin
  Temp:= TSquareMatrix.Create(Fn-1);
  fi:=0;
  For ei:=1 to Fm-1 do
  begin
    if ei=i then fi:=1;
    fj:=0;
    For ej:=1 to Fn-1 do
    begin
      if ej=j then fj:=1;
      Temp.Value[ei,ej]:=Value[ei+fi,ej+fj];
    end;
  end;
  Result:=Temp;
end;

Function TSquareMatrix.CoFactor(i, j: byte): Float;
begin
  if ((i+j) mod 2) = 0
  then result:=  minor(i,j).determinant
  else result:= -minor(i,j).determinant;
end;

Function TSquareMatrix.Determinant: Float;
var j: byte;
    Sum:Float;
begin
  if Fn=1 then Result:=Value[1,1]
  else
  begin
    Sum:=0;
    For j:=1 to Fn do Sum:= Sum+ Value[1,j]*CoFactor(1,j);
    Result:=Sum
  end;
end;

Function TCustomTrend.Total: Float;
var t: Word;
    Sum: Float;
begin
  Sum := 0;
  for t := Low(FArr) to High(FArr)
    do Sum := Sum + FArr[t];
  Result:=Sum;
end;

function TCustomTrend.Count: Integer;
begin
  Result:=Length(FArr);
end;

function TCustomTrend.Sum_t: integer;
begin
  Result:= (n* (n+1)) div 2;
  // Epsi Note:
  // sum of t*t has better precision than sum of sqr(t)
  // But equation is best method to do sum
end;

function TCustomTrend.Sum_t2: integer;
begin
  Result:= (n* (n+1)* (2*n+1)) div 6;
end;


Function TCustomTrend.StandardDeviation: Float;
var t: Word;
    Sum: Float;
begin
  Sum := 0;
  for t := 0 to High(FArr) do
    Sum := Sum + Sqr(FArr[t] - GetValue(t));
  Result:= Sqrt( Sum/(n-f));
end;

Constructor TConstantTrend.Create(Var AnArr: Array of Float);
var t: Word;
begin
  SetLength(FArr, Length(AnArr));
  for t := 0 to High(AnArr) do FArr[t] := AnArr[t];
  f:=1;
  a:=Total/Count;
end;

Function TConstantTrend.Equation: String;
begin
  Result:='Y = '+ FloatToStrF(a, ffFixed, 7,3);
end;

Function TConstantTrend.GetValue(t: integer): Float;
begin
  Result:=a;
end;

Constructor TLinearTrend.Create(Var AnArr: Array of Float);
var t: Word;
    Sum_d, Sum_dt: Float;
    det_d, det_1, det_t: Float;
    Matrix: TSquareMatrix;
begin
  SetLength(FArr, Length(AnArr));
  for t := 0 to High(AnArr) do FArr[t] := AnArr[t];
  f:=2;

  Sum_d:=0; Sum_dt:=0;
  for t := 1 to Length(FArr) do
  begin
    Sum_d  := Sum_d  + FArr[t-1];
    Sum_dt := Sum_dt + FArr[t-1]*t;
  end;

  Matrix:= TSquareMatrix.Create(f+1);
  Matrix.AssignFloat( [0,      0,     0,
                       Sum_d,  n,     Sum_t,
                       Sum_dt, Sum_t, Sum_t2] );
  Det_d:= Matrix.CoFactor(1,1);
  Det_1:= Matrix.CoFactor(1,2);
  Det_t:= Matrix.CoFactor(1,3);
  Matrix.Free;

  a:= -det_1/det_d;
  b:= -det_t/det_d;
end;

Function TLinearTrend.Equation: String;
begin
  Result:='Y = '+ FloatToStrF(a, ffFixed, 7,3) +' '+FToStr(b)+'*t';
end;

Function TLinearTrend.GetValue(t: integer): Float;
begin
  Result:=a + b*t;
end;

Procedure TCustomCyclicTrend.SetNumberOfCyclic(NewNumberOfCyclic:integer);
begin
  nc:=NewNumberOfCyclic
end;

Constructor TCyclicTrend.Create
            (Var AnArr: Array of Float; Const NumberOfCyclic : integer);
var t : Word;
    Matrix: TSquareMatrix;
    Sum_d, sum_dcos, sum_dsin: Float;
    det_d, det_1, det_cos, det_sin: Float;
begin
  SetLength(FArr, Length(AnArr));
  for t := 0 to High(AnArr) do FArr[t] := AnArr[t];
  f:=3;
  nc:=NumberOfCyclic;

  Sum_d:=0; Sum_dcos:=0; Sum_dsin:=0;
  for t := 1 to Length(FArr) do
  begin
    Sum_d    := Sum_d    + FArr[t-1];
    Sum_dcos := Sum_dcos + FArr[t-1]*cos(2*pi/nc*t);
    Sum_dsin := Sum_dsin + FArr[t-1]*sin(2*pi/nc*t);
  end;

  Matrix:= TSquareMatrix.Create(f+1);
  Matrix.AssignFloat( [0,        0,   0,   0,
                       Sum_d,    n,   0,   0,
                       Sum_dcos, 0,   n/2, 0,
                       Sum_dsin, 0,   0,   n/2] );
  Det_d:=   Matrix.CoFactor(1,1);
  Det_1:=   Matrix.CoFactor(1,2);
  Det_cos:= Matrix.CoFactor(1,3);
  Det_sin:= Matrix.CoFactor(1,4);
  Matrix.Free;

  a:= -det_1/det_d;
  u:= -det_cos/det_d;
  v:= -det_sin/det_d;
end;

Function TCyclicTrend.Equation: String;
begin
  Result:='Y = '+ FloatToStrF(a, ffFixed, 7,3) +' '
                + FToStr(u)+'*cos(2*pi*t/'+IntToStr(nc)+') '
                + FToStr(v)+'*sin(2*pi*t/'+IntToStr(nc)+')';
end;

Function TCyclicTrend.GetValue(t: integer): Float;
begin
  Result:=a + u*cos(2*pi*t/nc) + v*cos(2*pi*t/nc);
end;

Constructor TLinearCyclicTrend.Create
            (Var AnArr: Array of Float; Const NumberOfCyclic : integer);
var t: Word;
    Matrix: TSquareMatrix;
    Sum_d, Sum_dt, Sum_dcos, Sum_dsin, Sum_tsin: Float;
    det_d, det_1, det_t, det_cos, det_sin: Float;
begin
  SetLength(FArr, Length(AnArr));
  for t := 0 to High(AnArr) do FArr[t] := AnArr[t];

  f:=4;
  nc:=NumberOfCyclic;

  Sum_d:=0; Sum_dt:=0; Sum_dcos:=0; Sum_dsin:=0; Sum_tsin:=0;
  for t := 1 to Length(FArr) do
  begin
    Sum_d    := Sum_d    + FArr[t-1];
    Sum_dt   := Sum_dt   + FArr[t-1]*t;
    Sum_dcos := Sum_dcos + FArr[t-1]*cos(2*pi/nc*t);
    Sum_dsin := Sum_dsin + FArr[t-1]*sin(2*pi/nc*t);
    Sum_tsin := Sum_tsin + t*sin(2*pi/nc*t);
  end;

  Matrix:= TSquareMatrix.Create(f+1);
  Matrix.AssignFloat( [0,        0,     0,        0,   0,
                       Sum_d,    n,     Sum_t,    0,   0,
                       Sum_dt,   Sum_t, Sum_t2,   n/2, Sum_tsin,
                       Sum_dcos, 0,     n/2,      n/2, 0,
                       Sum_dsin, 0,     Sum_tsin, 0,   n/2] );
  Det_d:=   Matrix.CoFactor(1,1);
  Det_1:=   Matrix.CoFactor(1,2);
  Det_t:=   Matrix.CoFactor(1,3);
  Det_cos:= Matrix.CoFactor(1,4);
  Det_sin:= Matrix.CoFactor(1,5);
  Matrix.Free;

  a:= -det_1/det_d;
  b:= -det_t/det_d;
  u:= -det_cos/det_d;
  v:= -det_sin/det_d;
end;

Function TLinearCyclicTrend.Equation: String;
begin
  Result:='Y = '+ FloatToStrF(a, ffFixed, 7,3) +' '
                + FToStr(b)+'*t '
                + FToStr(u)+'*cos(2*pi*t/'+FToStr(nc)+') '
                + FToStr(v)+'*sin(2*pi*t/'+FToStr(nc)+')';
end;

Function TLinearCyclicTrend.GetValue(t: integer): Float;
begin
  Result:=a + b*t + u*cos(2*pi*t/nc) + v*cos(2*pi*t/nc);
end;

end.
